using JuMP
using LinearAlgebra
using DataFrames
using Gurobi
using CSV

OPTIMIZER = optimizer_with_attributes(Gurobi.Optimizer,
    "Method" => 2,
    "BarConvTol" => 1e-5,
    "Crossover" => 0)
    
sp=Model(OPTIMIZER)


#read in price data, append to beginning 12 hour period. Prices in $/MWh
prices_df = CSV.read(joinpath(pwd(), "NCA.csv"), DataFrame, types=[Float64], header=false)
prices_arr = vec(Matrix(prices_df))
#prices_arr = [18, 13, 13, 16, 20, 33, 45, 30, 6, 2, 0, -2, -5, -7, -6, -5, 0, 6, 35, 52, 51, 43, 35, 25]
#prices_arr = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500,]
pPrice = append!(ones(10), prices_arr)

#deleteat!(pPrice, 517:8772)


# last timestep in time set 1:T, hours
T = 8770

#maximum injection bottomhole pressure, psi
pMaxIPressure = 5317

#minimum/baseline injection bottomhole pressure, psi
pMinIPressure = 4592

#minimum/baseline production bottomhole pressure
pMinPPressure = 4153

#Injection pump total efficiency
pPeff = 0.875

#Combined conversion factor for STB/Min -> m^3/s and MPa -> psi
pConv = 0.00265*0.006895

#MW/psi of parasitic load for the injection pump at 23.8 STB/min
pInjectionPowerFactor = pConv/pPeff

#gives plant gross power output (MW) per unit production flow rate (STB/min)
pProductionPowerFactor = 0.1860

#Additional  injection pressure due to well friction at incremental injection rates
pWellFric2 = 171
pWellFric3 = 456
pWellFric4 = 856
pWellFric5 = 1369
pWellFric6 = 1996


#hydrostatic pressure at injection well bottomhole
pHydrostatic = 3796


#COSTING

#capital cost of generator per MW
pGENCOST = 2.17e6*(T-10)/8760

#capital cost of pump per MW
pPUMPCOST = 4.2e5*(T-10)/8760

#capital cost of transmission lines per MW
pTRANSCOST = 1.3e5*(T-10)/8760

#Annuity for plant oversizing
pANN = 0.051 + 0.018 + 0.0075
#annuity for interconnection oversizing
pCRFT = 0.0433

#baseline plant gross capacity (MW)
pCAP = 11.10

#baseline pump capacity
pPUMPCAP = 796*60*pInjectionPowerFactor

#baseline interconnection capacity
pICAP = pCAP - pPUMPCAP

#fraction of plant cost scaling with flow rate
pExtraFrac = 0.8

#amount of extra flow that the turbine can handle above nominal
pPeaking = 1.1

@variables(sp, begin
    #control variables
    #max power capacity (generator size)
    vPCAP >= pCAP
    
    #max power capacity (interconnection)
    vICAP >= pICAP
    
    #pump power capacity
    vPUMPCAP >= pPUMPCAP
    
    #control variable, flow from well
    0 <= vPROD[t = 1:T] <= 200
    #Injection pumping can be increased or decreased in increments
    20 >= vPUP1[t = 1:T] >= 0
    20 >= vPUP2[t = 1:T] >= 0
    20 >= vPUP3[t = 1:T] >= 0
    20 >= vPDN1[t = 1:T] >= 0
    20 >= vPDN2[t = 1:T] >= 0
    20 >= vPDN3[t = 1:T] >= 0
    #state variables
    # Injection BHP
    pMaxIPressure >= vIPRESSURE[t = 1:T] >= pMinIPressure
    #Difference betwen Injection BHP and production BHP
    vPPRESSURE[t = 1:T] >= pMinPPressure
end)

#for t in 1:T
#    setvalue(vPUMPON[t,1], 1)
#end
#expression for total cost of generation
@expression(sp, eGENCOST, vPCAP*pGENCOST*pANN*pExtraFrac + vPUMPCAP*pPUMPCOST*pANN + vICAP*pTRANSCOST*pCRFT)

#injection powers dependent on pump on/off
@expression(sp, ePUMPPOWER[t = 1:T], (vIPRESSURE[t] - pHydrostatic)*60*pInjectionPowerFactor +
                                       (1521 + 4*pWellFric4 - 3*pWellFric3)*pInjectionPowerFactor*vPUP1[t] +
                                       (1521 + 5*pWellFric5 - 4*pWellFric4)*pInjectionPowerFactor*vPUP2[t] +
                                       (1521 + 6*pWellFric6 - 5*pWellFric5)*pInjectionPowerFactor*vPUP3[t] -
                                       (340 + 3*pWellFric3 - 2*pWellFric2)*pInjectionPowerFactor*vPDN1[t] -
                                       (340 + 2*pWellFric2)*pInjectionPowerFactor*vPDN2[t] -
                                       340*pInjectionPowerFactor*vPDN3[t]
                                      )
                                      
#production power at any timestep
@expression(sp, eP[t = 1:T],  pProductionPowerFactor*vPROD[t])

#total revenue at any timestep
@expression(sp, eNET[t = 1:T], (eP[t]-ePUMPPOWER[t])*pPrice[t])


#expression for total injection rate
@expression(sp, eINJ[t = 1:T], 60 + vPUP1[t] + vPUP2[t] + vPUP3[t] - vPDN1[t] - vPDN2[t] - vPDN3[t])

#constraints requiring pumping levels to activate in the correct order
@constraint(sp, [t = 1:T], vPUP2[t] <= vPUP1[t])
@constraint(sp, [t = 1:T], vPUP3[t] <= vPUP2[t])
@constraint(sp, [t = 1:T], vPDN2[t] <= vPDN1[t])
@constraint(sp, [t = 1:T], vPDN3[t] <= vPDN2[t])

#constraints preventing simultaneous increases and reductions in pumping
@constraint(sp, [t = 1:T], vPUP1[t] + vPDN1[t] <= 20)
@constraint(sp, [t = 1:T], vPUP1[t] + vPDN2[t] <= 20)
@constraint(sp, [t = 1:T], vPUP1[t] + vPDN3[t] <= 20)
@constraint(sp, [t = 1:T], vPUP2[t] + vPDN1[t] <= 20)
@constraint(sp, [t = 1:T], vPUP2[t] + vPDN2[t] <= 20)
@constraint(sp, [t = 1:T], vPUP2[t] + vPDN3[t] <= 20)
@constraint(sp, [t = 1:T], vPUP3[t] + vPDN1[t] <= 20)
@constraint(sp, [t = 1:T], vPUP3[t] + vPDN2[t] <= 20)
@constraint(sp, [t = 1:T], vPUP3[t] + vPDN3[t] <= 20)


#total pumping power constrained by pump capacity
@constraint(sp, [t = 1:T], ePUMPPOWER[t] <= vPUMPCAP)

#objective function: total revenue minus cost of surface facilities oversizing
@objective(sp, Max, sum(eNET[t] for t in 11:T) - (eGENCOST-(pCAP*pGENCOST*pANN*pExtraFrac + pPUMPCAP*pPUMPCOST*pANN + pICAP*pTRANSCOST*pCRFT))) 


#power capacity constraints
@constraint(sp, [t = 1:T], eP[t] <= pPeaking*vPCAP)
@constraint(sp, [t = 1:T], (eP[t]-ePUMPPOWER[t]) <= vICAP)
@constraint(sp, [t = 1:T], (ePUMPPOWER[t]-eP[t]) <= vICAP)


#total injection constraint
@constraint(sp, [t = 1:T], (sum(eINJ[11:T]/(T-10)) <= 60))
#flow rate-of change constraint
#@constraint(sp, [t = 11:T], vPROD[t] <= vPROD[t-1]+14)

#initial conditions
@constraint(sp, [t = 1:10], vPROD[t] == 59.67)
@constraint(sp, [t = 1:10], vPPRESSURE[t] == pMinPPressure)
@constraint(sp, [t = 1:10], vIPRESSURE[t] == pMinIPressure)
@constraint(sp, [t = 1:10], vPUP1[t] == 0)
@constraint(sp, [t = 1:10], vPUP2[t] == 0)
@constraint(sp, [t = 1:10], vPUP3[t] == 0)
@constraint(sp, [t = 1:10], vPDN1[t] == 0)
@constraint(sp, [t = 1:10], vPDN2[t] == 0)
@constraint(sp, [t = 1:10], vPDN3[t] == 0)   

#expressions for longer, later sections of linearized well functions
@expression(sp, eP25[t = 11:T], sum(vPROD[s]-vPROD[s-1] + (eINJ[s-1]-eINJ[s]) for s in intersect(((t-25):(t-10)),(2:T))))
@expression(sp, eP50[t = 11:T], sum(vPROD[s]-vPROD[s-1] + (eINJ[s-1]-eINJ[s]) for s in intersect(((t-50):(t-26)),(2:T))))
@expression(sp, eP100[t = 11:T], sum(vPROD[s]-vPROD[s-1] + (eINJ[s-1]-eINJ[s]) for s in intersect(((t-100):(t-51)),(2:T))))
#@expression(sp, eP150[t = 11:T], sum(vPROD[s]-vPROD[s-1] + (eINJ[s-1]-eINJ[s]) for s in intersect(((t-150):(t-101)),(2:T))))
@expression(sp, eP200[t = 11:T], sum(vPROD[s]-vPROD[s-1] + (eINJ[s-1]-eINJ[s]) for s in intersect(((t-200):(t-101)),(2:T))))
#@expression(sp, eP250[t = 11:T], sum(vPROD[s]-vPROD[s-1] + (eINJ[s-1]-eINJ[s]) for s in intersect(((t-250):(t-201)),(2:T))))

#well function late-time linear/steady-state coefficient
pSS = 0.044
#model has control
for t in 11:T
   #production pressure depends on changes in flows in previous 200 timesteps multiplied by the coefficients of the production well functions
   @constraint(sp, vPPRESSURE[t] == vPPRESSURE[t-1] - (vPROD[t-1]+0.33-eINJ[t-1])*pSS  - (vPROD[t] - vPROD[t-1])*(0.902-pSS) -
                                  (vPROD[t-1]-vPROD[t-2])*(0.753-pSS) - (eINJ[t-2]-eINJ[t-1])*(0.974-pSS) -
                                  (vPROD[t-2]-vPROD[t-3])*(0.441-pSS) - (eINJ[t-3]-eINJ[t-2])*(0.467-pSS) -
                                  (vPROD[t-3]-vPROD[t-4])*(0.340-pSS) - (eINJ[t-4]-eINJ[t-3])*(0.355-pSS) -
                                  (vPROD[t-4]-vPROD[t-5])*(0.286-pSS) - (eINJ[t-5]-eINJ[t-4])*(0.294-pSS) -
                                  (vPROD[t-5]-vPROD[t-6])*(0.253-pSS) - (eINJ[t-6]-eINJ[t-5])*(0.260-pSS) -
                                  (vPROD[t-6]-vPROD[t-7])*(0.230-pSS) - (eINJ[t-7]-eINJ[t-6])*(0.236-pSS) -
                                  (vPROD[t-7]-vPROD[t-8])*(0.215-pSS) - (eINJ[t-8]-eINJ[t-7])*(0.214-pSS) -
                                  (vPROD[t-8]-vPROD[t-9])*(0.195-pSS) - (eINJ[t-9]-eINJ[t-8])*(0.197-pSS) -
                                  (vPROD[t-9]-vPROD[t-10])*(0.187-pSS) - (eINJ[t-10]-eINJ[t-9])*(0.192-pSS) - 
                                  eP25[t]*(0.135-pSS) - eP50[t]*(0.085-pSS) - eP100[t]*(0.058-pSS) - eP200[t]*(0.048-pSS)
                                 )

   #injection pressure depends on changes in flows in previous 200 timesteps multiplied by the coefficients of the injection well functions
   @constraint(sp, vIPRESSURE[t] == vIPRESSURE[t-1] - (vPROD[t-1]+0.33-eINJ[t-1])*pSS  - (eINJ[t-1] - eINJ[t])*(2.556-pSS) -
                                  (vPROD[t-1]-vPROD[t-2])*(0.957-pSS) - (eINJ[t-2]-eINJ[t-1])*(0.761-pSS) -
                                  (vPROD[t-2]-vPROD[t-3])*(0.461-pSS) - (eINJ[t-3]-eINJ[t-2])*(0.447-pSS) -
                                  (vPROD[t-3]-vPROD[t-4])*(0.352-pSS) - (eINJ[t-4]-eINJ[t-3])*(0.343-pSS) -
                                  (vPROD[t-4]-vPROD[t-5])*(0.295-pSS) - (eINJ[t-5]-eINJ[t-4])*(0.287-pSS) -
                                  (vPROD[t-5]-vPROD[t-6])*(0.253-pSS) - (eINJ[t-6]-eINJ[t-5])*(0.260-pSS) -
                                  (vPROD[t-6]-vPROD[t-7])*(0.230-pSS) - (eINJ[t-7]-eINJ[t-6])*(0.236-pSS) -
                                  (vPROD[t-7]-vPROD[t-8])*(0.215-pSS) - (eINJ[t-8]-eINJ[t-7])*(0.214-pSS) -
                                  (vPROD[t-8]-vPROD[t-9])*(0.195-pSS) - (eINJ[t-9]-eINJ[t-8])*(0.197-pSS) -
                                  (vPROD[t-9]-vPROD[t-10])*(0.187-pSS) - (eINJ[t-10]-eINJ[t-9])*(0.192-pSS) - 
                                  eP25[t]*(0.135-pSS) - eP50[t]*(0.085-pSS) - eP100[t]*(0.058-pSS) - eP200[t]*(0.048-pSS)
                                 )



       #flow constrained from above by current reservoir pressure
    @constraint(sp, vPROD[t] <= 0.1694*vPPRESSURE[t]-643.8)

end

optimize!(sp)

println("Hourly Revenue = ", objective_value(sp)/(T-10)/(pCAP-796*60*pInjectionPowerFactor))
println("Revenue Improvement = ", (objective_value(sp)/(T-10)/(pCAP-796*60*pInjectionPowerFactor)/(sum(prices_arr)/8760)) - 1)
println("Plant Oversizing = ",value.(vPCAP)/pCAP)
println("Pump Oversizing = ", value.(vPUMPCAP)/pPUMPCAP)
println("Interconnection Oversizing = ", value.(vICAP)/pICAP)




optimal_vPROD = value.(vPROD).*2.65
optimal_Inj = value.(eINJ).*2.65
optimal_vPPRESSURE = value.(vPPRESSURE).*0.00689476
optimal_vIPRESSURE = value.(vIPRESSURE).*0.00689476
optimal_NETREV = value.(eNET)
normgen = value.(eP)./(pCAP-796*60*pInjectionPowerFactor)
normld = value.(ePUMPPOWER)./(pCAP-796*60*pInjectionPowerFactor)



#remove first 12 dummy timesteps
deleteat!(optimal_vPROD, 1:10)
deleteat!(optimal_Inj, 1:10)
deleteat!(optimal_vPPRESSURE, 1:10)
deleteat!(optimal_vIPRESSURE, 1:10)
deleteat!(optimal_NETREV, 1:10)
deleteat!(normgen, 1:10)
deleteat!(normld, 1:10)

chg = zeros(8760)
dis = zeros(8760)


for i in 1:8760
    if (normgen[i]-normld[i]) - 1 < 0
    chg[i] = 1 - (normgen[i]-normld[i])
    end
end

for j in 1:8760
    if (normgen[j]-normld[j]) - 1 >= 0
    dis[j] = normgen[j]-normld[j] - 1
    end
end

net = zeros(8760)

for k in 1:8760
    net[k] = normgen[k] - normld[k]
end

println("Objective = ", objective_value(sp))

println("Gen loss =", sum(net)/8760)
println("Storage Efficiency = ", sum(dis)/sum(chg))

df = DataFrame(TIME = collect(1:(T-10)), FLOW = optimal_vPROD, INJECTION = optimal_Inj, PRODUCTION_BHP = optimal_vPPRESSURE, INJECTION_BHP = optimal_vIPRESSURE, REVENUE = optimal_NETREV, GEN = normgen, LOAD = normld)
CSV.write(joinpath(pwd(), "Optimal.csv"), df)